import typing

from networkx.utils import open_file
import json


@open_file(0, mode="r")
def read_instance(path) -> typing.Dict:
    data = json.load(path)
    if data["type"] != "CGSHOP2023_Instance":
        raise ValueError("Not a CGSHOP2023 instance file")
    if not data["name"] or not isinstance(data["name"], str):
        raise ValueError("Missing instance name")
    if len(data["outer_boundary"]) < 3:
        raise ValueError("Outer boundary must have at least three points!")
    return data


@open_file(0, mode="r")
def read_solution(path) -> typing.Dict:
    data = json.load(path)
    return parse_solution(data)


class NoSolution(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class BadSolutionFile(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


def parse_solution(data):
    if data["type"] != "CGSHOP2023_Solution":
        raise NoSolution("Not a CGSHOP2023 solution file")
    if "id" in data and "instance" not in data:
        data["instance"] = data["id"]
    if "name" in data and "instance" not in data:
        data["instance"] = data["name"]
    if not data["instance"] or not isinstance(data["instance"], str):
        raise BadSolutionFile("Missing instance name")
    data["instance"] = data["instance"].split("/")[-1].split(".")[0]
    polygons = data["polygons"]
    if not isinstance(polygons, list):
        raise BadSolutionFile("Solution is not a list.")
    data["polygons"] = [p for p in data["polygons"] if p]  # remove empty polygons
    if not polygons:
        raise BadSolutionFile("At least one polygon must be provided")
    if not all(isinstance(p, list) for p in polygons):
        raise BadSolutionFile("Badly encoded polygon. All polygons need to be lists.")
    if not all(len(p) >= 3 for p in polygons):
        raise BadSolutionFile(
            "All polygons need to consist of at least three distinct points."
        )
    return data



import os
import abc
import typing

from ..io import read_instance


class InstanceBaseDatabase(abc.ABC):
    """
    This class is only an ABC for lower level access classes.
    """

    def __init__(self, path: str, enable_cache: bool = False):
        """
        Create an InstanceDatabase that searches in a specified folder for instances.
        :param path: Path to the folder that contains the instance files (e.g. the folder
                        that contains the extracted zips). The instance files can be
                        in subfolders but have the names have to be NAME.instance.json.
        :param enable_cache: Should the loaded instances be cached? This can take quite
                        a lot of memory
        """
        self._path = path
        self._is_cache_enabled = enable_cache
        self._cache = {}
        if not os.path.exists(path):
            raise ValueError(f"The folder {os.path.abspath(path)} does not exist")

    def _filename_fits_name(self, filename, name):
        if not self._filename_fits_instance_convention(filename):
            return False
        split = filename.split(".")
        return split[0] == name

    def _filename_fits_instance_convention(self, filename):
        split = filename.split(".")
        if len(split) != 3:
            return False
        return split[1] == "instance" and split[2] == "json"

    def read(self, f):
        return read_instance(f)

    def _cache_and_return(self, instance):
        if self._is_cache_enabled:
            self._cache[instance.name] = instance
        return instance

    def _is_hidden_folder_name(self, name):
        if name.replace(".", "") and name[0] == ".":  # classic hidden unix files.
            return True
        if len(name) > 1 and name[:2] == "__":  # OS X does that.
            return True
        return False

    def _is_hidden_folder(self, path):
        return any(self._is_hidden_folder_name(folder) for folder in path.split("/"))

    def _extract_instance_name_from_path(self, path):
        filename = os.path.split(path)[1]
        return filename.split(".")[0]

    @abc.abstractmethod
    def __iter__(self) -> typing.Dict:
        """
        Iterate over all instances in database.
        :return: Instance objects
        """
        pass

    @abc.abstractmethod
    def __getitem__(self, name: str) -> typing.Dict:
        """
        Returns the instance of a specific name or throws an KeyError.
        :param name: Name of the instance.
        :return: Instance object
        """
        pass



"""lossless conversion to native format"""
from ..core import (
    FieldNumber,
    Point,
    Polygon,
    PolygonWithHoles,
)


def _str_to_number(number_data: str) -> FieldNumber:
    number_data = number_data.strip()
    if "." in number_data:
        elements = number_data.split(".")
        if len(elements) != 2:
            raise ValueError(f"Cannot parse {number_data}.")
        x = _str_to_number(elements[0]) + (
            _str_to_number(elements[1]) / _to_number(10 ** len(elements[1]))
        )
        return x
    while len(number_data) > 1 and number_data[0] == "0":  # remove leading zeros
        number_data = number_data[1:]
    return FieldNumber(number_data)


def _to_number(number_data):
    if isinstance(number_data, float):
        if int(number_data) == number_data:
            return _to_number(int(number_data))
        try:
            num, fp = str(number_data).split(".")
            len_fp = len(str(number_data).split(".")[1])
            x = _to_number(num) + (_to_number(fp) / _to_number(10**len_fp))
            return x
        except Exception:
            pass
        raise ValueError("Floating point not supported.")
    if isinstance(number_data, int):
        if number_data < (2**60):  # large ints in python are special...
            return FieldNumber(number_data)
        else:
            return _str_to_number(str(number_data))
    if isinstance(number_data, str):
        return _str_to_number(number_data)
    if isinstance(number_data, dict):
        return _to_number(number_data["num"]) / _to_number(number_data.get("den", 1))
    raise ValueError(f"Don't know how to convert '{number_data}'.")


def _to_coordinate(point_data):
    if "x" not in point_data or "y" not in point_data:
        raise ValueError(f"Cannot parse point data '{point_data}'")
    x = _to_number(point_data["x"])
    y = _to_number(point_data["y"])
    return Point(x, y)


def _to_polygon(points_data):
    return Polygon([_to_coordinate(p) for p in points_data])


def _to_polygon_with_holes(boundary, holes):
    boundary = _to_polygon(boundary)
    if not float(boundary.area()) > 0:
        raise ValueError("Polygon with negative boundary volume.")
    holes = [_to_polygon(hole) for hole in holes]
    if not all(float(hole.area()) < 0 for hole in holes):
        raise ValueError("Polygon has clockwise holes.")
    return PolygonWithHoles(boundary, holes)



"""
Just simple parsing of json files, including some checking and cleaning.
"""
# flake8: noqa F401
from .read import (
    read_instance,
    read_solution,
    parse_solution,
    BadSolutionFile,
    NoSolution,
)



"""
This file imports from the native code. This helps with coding, as the native part
is not available during coding but only after compilation.
"""
# flake8: noqa F401
from ._cgshop2023_core import (
    FieldNumber,
    Point,
    Polygon,
    PolygonWithHoles,
    NativeInstance,
    NativeSolution,
    area,
    verify,
    verify_instance,
)  # will only be available after building.



"""
A Python module for parsing and verifying instances and solutions for the
CG:SHOP 2023 challenge.

Use `read_instance(path)->dict` to parse an instance.
Use `read_solution(path)->dict` to parse a solution.

The InstanceDatabase is for reading the instances directly from the ZIP file
by name/instance-id. It can also cache the instances.

Use `verify(Instance, Solution)->str` to verify a solution.
It will return an empty string if everything is ok, otherwise a message
describing the problem.

This library uses a compiled C++-core. If you get segmentation faults, you
may want remove and reinstall it, in order to trigger a recompilation.
See the readme for further information.
"""
# flake8: noqa F401
from .io import read_solution, read_instance
from .instance_database import InstanceDatabase
from .verifier import verify



import os
import zipfile
import typing

from .instance_file_database import InstanceFileDatabase
from .instance_zip_database import InstanceZipDatabase


class InstanceDatabase:
    """
    This class allows to easily read instances from a folder/zipfile if the instance files
    follow the naming convention 'instance-name.instance.json'. It allows subfolder
    but no symbolic links.
    """

    def __init__(self, path: str, enable_cache: bool = False):
        """
        Create an InstanceDatabase that searches in a specified folder/zipfile for instances.
        :param path: Path to the folder/zipfile that contains the instance files (e.g. the folder
                        that contains the extracted zips). The instance files can be
                        in subfolders but have the names have to be NAME.instance.json.
        :param enable_cache: Should the loaded instances be cached? This can take quite
                        a lot of memory
        """
        self._inner_database = self._guess_database_class(path, enable_cache)

    def _guess_database_class(self, path: str, enable_cache):
        """
        Guess if the path contains a zipfile or a folder that could contain the database
        :param path: Path to the folder/zipfile
        :param enable_cache: Cache instances or read from file/zip
        :return: Guessed database object
        """
        if os.path.isdir(path):
            return InstanceFileDatabase(path, enable_cache=enable_cache)
        elif os.path.isfile(path):
            if zipfile.is_zipfile(path):
                return InstanceZipDatabase(path, enable_cache=enable_cache)
            else:
                raise FileNotFoundError(f"{path} is neither a directory or a zipfile.")
        raise FileNotFoundError(f"{path} not found")

    def __iter__(self) -> typing.Dict:
        """
        Iterate over all instances in database.
        :return: Instance objects
        """
        yield from self._inner_database

    def __getitem__(self, name: str) -> typing.Dict:
        """
        Returns the instance of a specific name or throws an KeyError.
        :param name: Name of the instance.
        :return: Instance object
        """
        if "/" in name:
            name = name.split("/")[-1]
        extension = ".instance"
        if len(name) > len(extension) and name[-len(extension) :] == extension:
            name = name[: -len(extension)]
        return self._inner_database[name]



"""
This file instructs scitkit-build how to build the module. This is very close
to the classical setuptools, but wraps some things for us to build the native
modules easier and more reliable.

For a proper installation with `pip install .`, you additionally need a
`pyproject.toml` to specify the dependencies to load this `setup.py`.

You can use `python3 setup.py install` to build and install the package
locally, with verbose output. To build this package in place, which may be
useful for debugging, use `python3 setup.py develop`. This will build
the native modules and move them into your source folder.

The setup options are documented here:
https://scikit-build.readthedocs.io/en/latest/usage.html#setup-options
"""

from setuptools import find_packages
from skbuild import setup
import sys


def run_conan():
    import subprocess

    # Make sure to access to local conan
    cmd = "-m conans.conan install . -if cmake --build=missing"
    subprocess.run([sys.executable, *cmd.split(" ")], check=True)


def readme():
    # Simply return the README.md as string
    with open("README.md") as file:
        return file.read()


run_conan()  # automatically running conan. Ugly workaround, but does its job.
setup(  # https://scikit-build.readthedocs.io/en/latest/usage.html#setup-options
    # ~~~~~~~~~ BASIC INFORMATION ~~~~~~~~~~~
    name="cgshop2023_pyutils",
    version="0.2.10",  # TODO: Use better approach for managing version number.
    description="Official utilities for the CG:SHOP Challenge 2023.",
    long_description=readme(),
    url="https://github.com/CG-SHOP/pyutils23",
    long_description_content_type="text/markdown",
    author="TU Braunschweig, IBR, Algorithms Group (Phillip Keldenich and Dominik Krupke)",
    author_email="keldenich@ibr.cs.tu-bs.de, krupke@ibr.cs.tu-bs.de",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    # ~~~~~~~~~~~~ CRITICAL PYTHON SETUP ~~~~~~~~~~~~~~~~~~~
    # This project structures defines the python packages in a subfolder.
    # Thus, we have to collect this subfolder and define it as root.
    packages=find_packages("python"),  # Include all packages in `./python`.
    package_dir={"": "python"},  # The root for our python package is in `./python`.
    python_requires=">=3.7",  # lowest python version supported.
    install_requires=[
        # requirements necessary for basic usage (subset of requirements.txt)
        "chardet>=4.0.0",
        "networkx>=2.5.1",
        "requests>=2.25.1",
    ],
    # ~~~~~~~~~~~ CRITICAL CMAKE SETUP ~~~~~~~~~~~~~~~~~~~~~
    # Especially LTS systems often have very old CMake version (or none at all).
    # Defining this will automatically install locally a working version.
    cmake_minimum_required_version="3.17",
    #
    # By default, the `install` target is built (automatically provided).
    # To compile a specific target, use the following line.
    # Alternatively, you can use `if(NOT SKBUILD) ... endif()` in CMake, to
    # remove unneeded parts for packaging (like tests).
    # cmake_install_target = "install"
    #
    # In the cmake you defined by install(...) where to move the built target.
    # This is critical als only targets with install will be used by skbuild.
    # This should be relative paths to the project root, as you don't know
    # where the package will be packaged. You can change the root for the
    # install-paths with the following line. Note that you can also access
    # the installation root (including this modification) in cmake via
    # `CMAKE_INSTALL_PREFIX`. If your package misses some binaries, you
    # probably messed something up here or in the `install(...)` path.
    # cmake_install_dir = ".",
    # |-----------------------------------------------------------------------|
    # | If you are packing foreign code/bindings, look out if they do install |
    # | targets in global paths, like /usr/libs/. This could be a problem.    |
    # |-----------------------------------------------------------------------|
    #
    # Some CMake-projects allow you to configure it using parameters. You
    # can specify them for this Python-package using the following line.
    # cmake_args=["-DCGALPY_KERNEL_BINDINGS=epec",
    #             "-DCGALPY_ARRANGEMENT_ON_SURFACE_2_BINDINGS=ON",]
    #
    # There are further options, but you should be fine with these above.
)



"""
This file contains the ZipSolutionIterator which can read a zip and return all solutions
in it. It should be reasonably robust and have some basic security features.
"""
import json
import typing
from typing import BinaryIO, Union, Iterator
from os import PathLike
from zipfile import ZipFile, BadZipFile
from json import JSONDecodeError

import chardet

from ..io import parse_solution, NoSolution, BadSolutionFile
from .zip_reader_errors import (
    BadZipChecker,
    NoSolutionsError,
    InvalidEncodingError,
    InvalidJSONError,
    InvalidZipError,
)


class ZipSolutionIterator:
    """
    Iterates over all solutions in a zip file.
    First initialize the class, then use call.
    e.g.,
    ```
    zsi = ZipSolutionIterator()
    for solution in zsi("./myzip.zip"):
        print(solution.instance_name)
    ```
    """

    def __init__(
        self,
        file_size_limit: int = 250 * 1_000_000,
        zip_size_limit: int = 2000 * 1_000_000,
        solution_extensions=("json", "solution"),
    ):
        """
        Set the parameters in the constructor. Use the __call__ to actually iterate
        a zip file.
        :param file_size_limit: Limit the size of a single file within the zip.
        :param zip_size_limit: Limit the overall decompressed size of the zip.
        :param solution_extensions: What file extensions should be checked?
        """
        self._checker = BadZipChecker(
            file_size_limit=file_size_limit, zip_size_limit=zip_size_limit
        )
        self._solution_extensions = solution_extensions

    def _check_if_bad_zip(self, zipfile):
        self._checker(zipfile)

    def _is_hidden_folder_name(self, name):
        if len(name) > 1 and name[0] == ".":  # classic hidden unix files.
            return True
        if len(name) > 1 and name[:2] == "__":  # OS X does that.
            return True
        return False

    def _is_solution_filename(self, name):
        extension = name.split(".")[-1].lower()
        # check extension
        if extension not in self._solution_extensions:
            return False
        # check for hidden file/directory
        if any(self._is_hidden_folder_name(s) for s in name.split("/")):
            return False
        # any more checks?
        return True

    def _iterate_solution_filenames(self, zip_file):
        had_filename = False
        for filename in zip_file.namelist():
            if self._is_solution_filename(filename):
                had_filename = True
                yield filename
        if not had_filename:
            raise NoSolutionsError()

    def _robust_parse_json_from_bytes(self, bytes):
        """
        First tries to encode via 'UTF-8', otherwise it uses chardet.
        :param bytes: bytes of solution file
        :return: json
        """
        try:
            return json.loads(str(bytes, encoding="utf-8", errors="strict"))
        except UnicodeDecodeError:
            encoding = chardet.detect(bytes)
            return json.loads(
                str(bytes, encoding=encoding["encoding"], errors="strict")
            )

    def _parse_file(self, solution_file, file_name, info):
        # read no more than the claimed file_size bytes (which we checked for limit violations)
        b = solution_file.read(info.file_size)
        try:
            return self._robust_parse_json_from_bytes(b)
        except UnicodeDecodeError as ude:
            raise InvalidEncodingError(file_name) from ude
        except JSONDecodeError as e:
            raise InvalidJSONError(file_name, f"{e}") from e
        except RecursionError as re:
            raise InvalidJSONError(file_name, "Nesting level is too deep") from re

    def _iterate_solution_jsons(self, zip_file):
        for file_name in self._iterate_solution_filenames(zip_file):
            with zip_file.open(file_name, "r") as sol_file:
                info = zip_file.getinfo(file_name)
                yield file_name, self._parse_file(sol_file, file_name, info)

    def __call__(
        self, path_or_file: Union[BinaryIO, str, PathLike]
    ) -> Iterator[typing.Dict]:
        """
        Iterates over all solutions in the zip.
        :param path_or_file: Zip or file
        :return:
        """
        found_an_instance = False
        try:
            with ZipFile(path_or_file) as zip_file:
                self._check_if_bad_zip(zip_file)
                for file_name, solution_json in self._iterate_solution_jsons(zip_file):
                    meta = {
                        "zip_info": {
                            "zip_file": zip_file.filename,
                            "file_in_zip": file_name,
                        }
                    }
                    try:
                        solution = parse_solution(solution_json)
                        solution["meta"] = meta
                        yield solution
                        found_an_instance = True
                    except NoSolution:
                        print(f"Skipping {file_name}, as it is not a solution file.")
        except BadZipFile as e:
            raise InvalidZipError(f"{e}") from e
        except BadSolutionFile as e:
            raise InvalidZipError(f"Aborted parsing zip due to bad file: {e}") from e
        if not found_an_instance:
            raise NoSolutionsError()



"""
This file provides functions to verify instances and solutions.
The primary logic is convering the data to a native format
with exact data types and wrapping the C++-code.
"""
# flake8: noqa F401
import typing

from ..core import (
    NativeInstance,
    NativeSolution,
    verify as verify_,
    verify_instance as _verify_instance,
)

from ._convert_to_native_format import _to_polygon, _to_polygon_with_holes


def verify(instance: typing.Dict, solution: typing.Dict):
    """
    Verify a solution for an instance. This function uses C++ code, CGAL, and exact arithmetics
    to obtain exact results within a few seconds.
    :param instance: The data of the instance as parsed from the json.
    :param solution: The data of the solution as parsed from the json. Use our parser
            to verify the correctness of the format.
    :return: An empty string if the solution is valid. Otherwise, an error message.
    """
    instance_poly = _to_polygon_with_holes(
        instance["outer_boundary"], instance["holes"]
    )
    n_instance = NativeInstance(instance_poly)
    solution_polys = [_to_polygon(poly) for poly in solution["polygons"]]

    n_solution = NativeSolution(solution_polys)
    if not all(float(p.area()) > 0 for p in solution_polys):
        return "Solution contains polygons of zero size."
    error_msg = verify_(n_instance, n_solution)
    return error_msg


def verify_instance(instance: typing.Dict):
    """
    Verify an instance to be valid.
    :param instance: The data of the instance as parsed from the json.
    :return: True if it is valid, otherwise false.
    """
    instance_poly = _to_polygon_with_holes(
        instance["outer_boundary"], instance["holes"]
    )
    n_instance = NativeInstance(instance_poly)
    return _verify_instance(n_instance)
